LOOP_WAIT = 1.0 / 200
BEATS_PER_BAR = 4
MIDI_EVENT_NOTE_ON = 0x90
NOTE_DURATIONS = [1, 1 / 2, 1 / 4, 1 / 8, 1 / 16]
DEFAULT_RATES = [1 / 2, 1 / 4, 1 / 8]
POSSIBLE_NOTES = [str(i) for i in range(128)]
